--|-----------------------------------------------------------------
--|-- G"unther K"ogel
--|-- University of Stuttgart
--|-----------------------------------------------------------------
-- Modula-2 sets based on bool_strings 

indexing
	date: "Tue Aug 11 16:57:53 MET DST 1992";
	revision: "1.2"

-- 1.2 new index range is 0 to MaxSize
-- 1.1 new index range is 0 to MaxSize - 1
-- 1.0 initial revision



class Sets
export
Union, Difference, Intersection, SymDiff, Complement, Include, Exclude,
Card, Size, Minimum, Maximum, Select, Extract, IsSubset, IsStrictSubset,
IsEqual, IsNotEqual, IsElement, IsEmpty, Assign, AssignElmt, AssignEmpty,
ReadSet, WriteSet, WriteElmt,
FirstElement {Sets}, LastElement {Sets}, Cardinal {Sets}, Set {Sets},
SetCardinal {Sets}, SetFirstElement {Sets}, SetLastElement {Sets}

feature

Set 		: BOOL_STRING;
FirstElement,
LastElement,
Cardinal	: INTEGER;
NoCard		: INTEGER IS -1;

Create (MaxSize : INTEGER) is
	require
			positive_size: MaxSize > 0
	do
		Set.Create (MaxSize + 1);
		Size := MaxSize + 1;
		AssignEmpty;
	end; -- Create

Union (other : like Current) is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	local
		index : INTEGER;
	do
		from
			index := 1
		until
			index >= Size
		loop
			if other.Set.item (index) then
				Set.put (TRUE, index);
			end;
			index := index + 1;
		end; -- loop
		Cardinal := NoCard; -- needs to be calculated
		FirstElement := Min (FirstElement, other.FirstElement); -- new bounderies
		LastElement  := Max (LastElement, other.LastElement);
	end; -- Union

Difference (other : like Current) is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	local
		index : INTEGER;
	do
		from
			index := 1
		until
			index >= Size
		loop
			if other.Set.item (index) then
				Set.put (FALSE, index);
			end;
			index := index + 1;
		end; -- loop
		Cardinal := NoCard;
	end; -- Difference

Intersection (other : like Current) is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	local
		index : INTEGER;
	do
		from
			index := 1
		until
			index >= Size
		loop
			if Set.item (index) and other.Set.item (index) then
			else
				Set.put (FALSE, index);
			end;
			index := index + 1;
		end; -- loop
		Cardinal := NoCard;
		FirstElement := Max (FirstElement, other.FirstElement);
		LastElement  := Min (LastElement,  other.LastElement);
	end; -- Intersection

SymDiff (other : like Current) is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	local
		index : INTEGER;
	do
		from
			index := 1
		until
			index >= Size
		loop
			if Set.item (index) xor other.Set.item (index) then
				Set.put (TRUE,  index);
			else
				Set.put (FALSE, index);
			end;
			index := index + 1;
		end; -- loop
		Cardinal := NoCard;
		FirstElement := Min (FirstElement, other.FirstElement);
		LastElement  := Max (LastElement,  other.LastElement);
	end; -- SymDiff

Complement is
	do
		Set := not Set;
		Cardinal     := NoCard;
		FirstElement := 1;
		LastElement  := Size;
	end; -- Complement

Include (Element_position : INTEGER) is
	require
		element_large_enough: Element_position >= 0;
		element_small_enough:Element_position <= Size
	local
		Position : INTEGER
	do
		Position := Element_position + 1;
		if Position > 0 and Position <= Size then
			Set.put (TRUE, Position);
			Cardinal := NoCard;
			if FirstElement > Position then
				FirstElement := Position;
			end;
			if LastElement < Position then
				LastElement := Position;
			end;
		end;
	end; -- Include

Exclude (Element_position : INTEGER) is
	require
		element_large_enough: Element_position >= 0;
		element_small_enough:Element_position <= Size	local
		Position : INTEGER
	do
		Position := Element_position + 1;
		if Position > 0 and Position <= Size then
			Set.put (FALSE, Position);
			Cardinal := NoCard;
			if Position = FirstElement and FirstElement /= Size then
				FirstElement := FirstElement + 1;
			end;
			if Position = LastElement and LastElement /= 1 then
				LastElement := LastElement - 1;
			end;
		end;
	end; -- Exclude

Card : INTEGER is
	local
		index,
		no_of_elements : INTEGER;
	do
		from
			index := FirstElement;
			no_of_elements := 0
		until
			index > LastElement
		loop
			if Set.item (index) then  
				no_of_elements := no_of_elements + 1;
			end;
			index := index + 1;
		end; -- loop
		RESULT := no_of_elements;
	ensure
		RESULT >= 0 and RESULT <= Size
	end;

Size : INTEGER;

Minimum : INTEGER is
	local
		index : INTEGER;
		quit_loop : BOOLEAN
	do
		from
			index := FirstElement;
			quit_loop := FALSE
		until
			index > LastElement or quit_loop
		loop
			if IsElement (index) then
				FirstElement := index;
				quit_loop := TRUE;
			end;
			index := index + 1;
		end; -- loop

		if quit_loop then
			RESULT := FirstElement;
		else
			RESULT := 0; -- empty
		end;
	ensure
		RESULT >= 0 and RESULT <= Size
		
	end; -- Minimum

Maximum : INTEGER is
	local
		index : INTEGER;
		quit_loop : BOOLEAN
	do
		from
			index := LastElement
		until
			index < FirstElement or quit_loop
		loop
			if IsElement (index) then
				LastElement := index;
				quit_loop := TRUE;
			end;
			index := index - 1;
		end; -- loop

		if quit_loop then
			RESULT := LastElement;
		else
			RESULT := 0; -- empty
		end;
	ensure
		RESULT >= 0 and RESULT <= Size		
	end; -- Maximum

Select : INTEGER is
	do
		RESULT := Minimum;
	ensure
		RESULT >= 0 and RESULT <= Size
	end; -- Select

Extract : INTEGER is
	local
		index : INTEGER;
	do
		index := Minimum;
		if index /= 0 then
			Exclude (index);
		end;
		RESULT := index;
	ensure
		RESULT >= 0 and RESULT <= Size
	end; -- Extract

IsSubset (other : like Current) : BOOLEAN is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	local
		index : INTEGER
	do
		from 
			index := 1
		until
			index > Size or else (Set.item (index) and not other.Set.item (index))
		loop
			index := index + 1;
		end; -- loop
		RESULT := index > Size;
	end; -- IsSubset

IsStrictSubset (other : like Current) : BOOLEAN is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	local
		index : INTEGER
	do
		RESULT := IsSubset (other) and IsNotEqual (other);
	end; -- IsStrictSubset

IsEqual (other : like Current) : BOOLEAN is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	local
		index : INTEGER
	do
		from 
			index := 1
		until
			index > Size or else (Set.item (index) /= other.Set.item (index))
		loop
			index := index + 1;
		end; -- loop
		RESULT := index > Size;
	end; -- IsEqual

IsNotEqual (other : like Current) : BOOLEAN is
	require
		other_void: not other.void;
		same_size : other.Size = Size
	do
		Result := not IsEqual (other);
	end; -- IsNotEqual

IsElement (Element_position : INTEGER) : BOOLEAN is
	require
		element_large_enough: Element_position >= 0;
		element_small_enough:Element_position <= Size
	local
		Position : INTEGER	
	do
		Position := Element_position + 1;
 		if Position > 0 and Position <= Size then
			RESULT := Set.item (Position);
		else
			RESULT := FALSE;
		end;
	end; -- IsElement

IsEmpty : BOOLEAN is
	local
		index : INTEGER;
	do
		if FirstElement <= LastElement then
			from
				index := FirstElement
			until
				index > LastElement or else Set.item (index)
			loop
				index := index + 1;
			end; -- loop
			RESULT := index > LastElement;
		else
			RESULT := TRUE;
		end;
		
	end; -- IsEmpty

-- Forall is -- not implemented

-- Exists is not implemented

-- Exists is not implemented

Assign : like Current is
	do
		RESULT := Duplicate;
	end; -- Assign

AssignElmt (Element_position : INTEGER) is
	require
		element_large_enough: Element_position >= 0;
		element_small_enough:Element_position <= Size
	local
		Position : INTEGER
	do
		Position := Element_position + 1;
 		if Position > 0 and Position <= Size then
			AssignEmpty;
			Include (Element_position);
			Cardinal := 1;
			FirstElement := Position;
			LastElement := Position;
		end;
	end; -- AssignElmt


AssignEmpty is
	do
		Set.all_false;
		Cardinal := 0;
		FirstElement := Size;
--		LastElement  := 0; old version
		LastElement  := 1;
	end; -- AssignEmpty

-- ForallDo is not implemented

ReadSet (f : FILE) is
	require
		              not f.void;
		not_open:     f.is_open_read;
--		not_readable: f.readable; -- not a feature of io.input
	local
		show   : FILE
	do
		if f.void then
			show := io.input;
		else
			show := f;
		end;
		show.unset_error_output; -- show no errors to user

	--	if show.readable and then show.is_open_read then -- see above
		if show.is_open_read then
			from	-- find start of set
				show.readchar
			until
				show.end_of_file or else show.lastchar = '{'
			loop
				show.readchar;
			end; -- loop

			from
			until
				show.end_of_file or else show.lastchar = '}' 
			loop
				show.readint;
				if show.error = 0 then -- got a number
					if show.lastint > 0 and show.lastint <= Size then -- if it fits
						Include (show.lastint);
					end;
				else
					show.readchar; -- maybe end of set mark '}'
				end;
			end; -- loop
		show.set_error_output; -- back to the default
		end; -- if
	end; -- ReadSet

WriteSet (f : FILE) is
	require
		not f.void
	local
		index : INTEGER;
		show   : FILE
	do
		if f.void then
			show := io.output;
		else
			show := f;
		end;
		show.putchar ('{');
		from
			index := FirstElement;
		until
			index > LastElement
		loop
			if IsElement (index) then
				WriteElmt (show, index);
			end;
			index := index + 1;
		end; -- loop
		show.putchar ('}');
		show.new_line;
	end; -- WriteSet

WriteElmt (f : FILE; Element_position : INTEGER) is
	require
		not f.void
	do
		f.putchar (' ');
		f.putint (Element_position);
	end;

Duplicate : like Current is
	local
		index : INTEGER;
	do
		RESULT.Create (Size);
		RESULT.Set.all_false;
		from
			index := FirstElement
		until 
			index > LastElement
		loop
			RESULT.Set.put (Set.item (index), index);
			index := index + 1;
		end; -- loop
		RESULT.SetCardinal (Cardinal);
		RESULT.SetFirstElement (FirstElement);
		RESULT.SetLastElement (LastElement);
	end; -- Duplicate

Min (a, b : INTEGER) : INTEGER is
	do
		if a < b then
			RESULT := a;
		else
			RESULT := b;
		end;
	end; -- Min

Max (a, b : INTEGER) : INTEGER is
	do
		if a < b then
			RESULT := b;
		else
			RESULT := a;
		end;
	end; -- Max

SetCardinal ( C : INTEGER) is
	do
		Cardinal := C;
	end; -- SetCardinal

SetFirstElement (F : INTEGER) is
	do
		FirstElement := F;
	end; -- SetFirstElement

SetLastElement (L : INTEGER) is
	do
		LastElement := L;
	end; -- SetLastElement

end; -- class Sets
