--|-----------------------------------------------------------------
--|-- G"unther K"ogel
--|-- University of Stuttgart
--|-----------------------------------------------------------------
-- simple interface for REX

indexing
	date: "Tue Jun  9 12:08:24 MET DST 1992";
	revision: "1.3"

-- 1.3 selective export
-- 1.2 add some assertions, maximum read = BUFFERSIZE - 2
-- 1.1 add ChBufferIndex, BytesRead, Eof, LineCount, LineStart
--     and procedures to set values
-- 1.0 inital revision

$@ class $source
export Open, Close, Read,
					-- data especially for REX
buffer,
ChBufferIndex,
BytesRead,
Eof,
LineCount,
LineStart,
SetChBufferIndex,
SetBytesRead,
SetEof,
SetLineCount,
SetLineStart
 
feature

BUFFERSIZE	: INTEGER is 2050;	-- default, the bigger the better
buffer		: STRING;		-- contains data read in
inputfile	: FILE;			-- source
read_from_stdin	: BOOLEAN;		-- no close necessary

					-- some data especially for REX
ChBufferIndex	: INTEGER;
BytesRead	: INTEGER;
Eof		: BOOLEAN;
LineCount	: INTEGER;
LineStart	: INTEGER;

Create  is
	do
		buffer.Create (BUFFERSIZE);		
	end;

Open (filename: STRING): BOOLEAN is
					-- if filename given then try to open file
					-- else connect with stdinput
	do
		if filename.void then
			inputfile := io.input;	-- when no name given, read from stdin
			read_from_stdin := TRUE;	-- prevent closeing of file
			Result := TRUE;
		else
			inputfile.Create (filename);
			if inputfile.exists then
				inputfile.open_read;
				Result := TRUE;
			else
				buffer.forget;
				Result := FALSE;	-- not useable
			end;
		end;
	end; -- Open

Close is
			-- if real file then close input file else leave untouched
	do
		if not read_from_stdin then
			inputfile.close;
			buffer.forget;
		end;
	end; -- Close

Read (wanted: INTEGER): INTEGER is
					-- fill buffer, return amount
	require
		positiv_and_not_zero: wanted > 0
	local
		amount	: INTEGER;
	do
		if wanted > BUFFERSIZE - 2 then
			amount := BUFFERSIZE - 2;	-- REX appends yyEobCh and '\0'
		elsif wanted < 1 then
			Result := 0;			 -- you get what you want
		else
			amount := wanted;
        end;
		inputfile.readstream (amount);
		buffer.append (inputfile.laststring);
		Result := inputfile.laststring.count;
	end; -- Read

SetChBufferIndex (value: INTEGER) is
	do
		ChBufferIndex := value;
	end; -- SetChBufferIndex

SetBytesRead (value: INTEGER) is
	do
		BytesRead := value;
	end; -- BytesRead

SetEof (value: BOOLEAN) is
	do
		Eof := value;
	end; -- SetEof

SetLineCount (value: INTEGER) is
	do
		LineCount := value;
	end; -- LineCount

SetLineStart (value: INTEGER) is
	do
		LineStart := value;
	end; -- LineStart

$@ end; -- $source
